/* WirePlumber
 *
 * Copyright © 2019 Collabora Ltd.
 *    @author George Kiagiadakis <george.kiagiadakis@collabora.com>
 *
 * SPDX-License-Identifier: MIT
 */

#include "error.h"

/*! \defgroup wperror Errors */

G_DEFINE_QUARK (wireplumber-library, wp_domain_library);
